/************************************************************
MPU9250_DMP_Quaternion
 Quaternion example for MPU-9250 DMP Arduino Library 
Jim Lindblom @ SparkFun Electronics
original creation date: November 23, 2016
https://github.com/sparkfun/SparkFun_MPU9250_DMP_Arduino_Library

The MPU-9250's digital motion processor (DMP) can calculate
four unit quaternions, which can be used to represent the
rotation of an object.

This exmaple demonstrates how to configure the DMP to 
calculate quaternions, and prints them out to the serial
monitor. It also calculates pitch, roll, and yaw from those
values.

Development environment specifics:
Arduino IDE 1.6.12
SparkFun 9DoF Razor IMU M0

Supported Platforms:
- ATSAMD21 (Arduino Zero, SparkFun SAMD21 Breakouts)
*************************************************************/
#include <SparkFunMPU9250-DMP.h>
#include <SPI.h>
#include "wiring_private.h" // pinPeripheral() function

// NRF Libraries
#include <nRF24L01.h>
#include <RF24.h>

#define SerialPort SerialUSB
#define NRF_ID 0

MPU9250_DMP imu;

RF24 radio(8, 10); // CE, CSN
const byte address[6] = "00001";

// NRF Packet structure for IMU-9250
// Maximum payload 32 Bytes
struct IMU_NRF_Data{ 
    char ID;
    float dRoll;
    float dPitch;
    float dYaw;
    float dAccelX;
    float dAccelY;
    float dAccelZ;
};
IMU_NRF_Data NRFPacket;

float myRoll, myPitch, myYaw;
float aux_myRoll, aux_myPitch, aux_myYaw;

void setup() 
{
  pinMode(10, OUTPUT); //CSN
  pinMode(8, OUTPUT); //CE

  //START NRF
  radio.begin();
  radio.openWritingPipe(address);
  radio.setPALevel(RF24_PA_MAX);
  radio.setDataRate(RF24_250KBPS);
  radio.stopListening();

  NRFPacket.ID = NRF_ID;
  
  SerialPort.begin(115200);

  // Call imu.begin() to verify communication and initialize
  if (imu.begin() != INV_SUCCESS)
  {
    while (1)
    {
      SerialPort.println("Unable to communicate with MPU-9250");
      SerialPort.println("Check connections, and try again.");
      SerialPort.println();
      delay(5000);
    }
  }

  // Enable all sensors:
  imu.setSensors(INV_XYZ_GYRO | INV_XYZ_ACCEL | INV_XYZ_COMPASS);
  // Accel options are +/- 2, 4, 8, or 16 g
  imu.setAccelFSR(2); // Set accel to +/-2g
  // setLPF() can be used to set the digital low-pass filter
  // of the accelerometer and gyroscope.
  // Can be any of the following: 188, 98, 42, 20, 10, 5
  // (values are in Hz).
  imu.setLPF(5); // Set LPF corner frequency to 5Hz
  
  imu.dmpBegin(DMP_FEATURE_6X_LP_QUAT | // Enable 6-axis quat
               DMP_FEATURE_SEND_RAW_ACCEL | //Send raw accelerometer values to FIFO
               DMP_FEATURE_GYRO_CAL | // Use gyro calibration
               DMP_FEATURE_SEND_CAL_GYRO, // Send Calibrated gyro values
              10); // Set DMP FIFO rate to 10 Hz
  // DMP_FEATURE_LP_QUAT can also be used. It uses the 
  // accelerometer in low-power mode to estimate quat's.
  // DMP_FEATURE_LP_QUAT and 6X_LP_QUAT are mutually exclusive
}

// Quaternions to Euler - GFM
void toEulerAngle_Roll_Yaw(float w, float x, float y, float z, float& myRoll, float& myPitch, float& myYaw)
{
  
  double ysqr = y * y;

  // roll (x-axis rotation)
  double t0 = +2.0 * (w * x + y * z);
  double t1 = +1.0 - 2.0 * (x * x + ysqr);
  myRoll = atan2(t0, t1) * (180.0 / PI);

  // pitch (y-axis rotation)
  double t2 = +2.0 * (w * y - z * x);
  t2 = t2 > 1.0 ? 1.0 : t2;
  t2 = t2 < -1.0 ? -1.0 : t2;
  myPitch = asin(t2) * (180.0 / PI);

  // yaw (z-axis rotation)
  double t3 = +2.0 * (w * z + x * y);
  double t4 = +1.0 - 2.0 * (ysqr + z * z);
  myYaw = atan2(t3, t4) * (180.0 / PI);

  if (myPitch < 0) myPitch = 360.0 + myPitch;
  if (myRoll < 0) myRoll = 360.0 + myRoll;
  if (myYaw < 0) myYaw = 360.0 + myYaw; 
}

void toEulerAngle_Pitch(float w, float x, float y, float z, float& myRoll, float& myPitch, float& myYaw)
{  
  double t0 = -2.0f * x * z + 2.0f * w * y;
  double t1 = z * z - y * y - x * x + w * w;
  double t2 = 2.0f * y * z + 2.0f * w * x;
  double t3 = -2.0f * x * y + 2.0f * w * z;
  double t4 = y * y - z * z - x * x + w * w;

  myPitch = atan2(t0, t1) * (180.0 / PI);
  myRoll = asin(t2) * (180.0 / PI);
  myYaw = atan2(t3, t4) * (180.0 / PI);

  if (myPitch < 0) myPitch = 360.0 + myPitch;
  if (myRoll < 0) myRoll = 360.0 + myRoll;
  if (myYaw < 0) myYaw = 360.0 + myYaw; 
}

void loop() 
{
  // Check for new data in the FIFO
  if ( imu.fifoAvailable() )
  {
    // Use dmpUpdateFifo to update the ax, gx, mx, etc. values
    if ( imu.dmpUpdateFifo() == INV_SUCCESS)
    {
      // computeEulerAngles can be used -- after updating the
      // quaternion values -- to estimate roll, pitch, and yaw
      //imu.computeEulerAngles();
      printIMUData();
    }
  }
}

void printIMUData(void)
{  
  float accelX = imu.calcAccel(imu.ax);
  float accelY = imu.calcAccel(imu.ay);
  float accelZ = imu.calcAccel(imu.az);

  toEulerAngle_Roll_Yaw(imu.calcQuat(imu.qw), imu.calcQuat(imu.qx), imu.calcQuat(imu.qy), imu.calcQuat(imu.qz), myRoll, aux_myPitch, myYaw);
  toEulerAngle_Pitch(imu.calcQuat(imu.qw), imu.calcQuat(imu.qx), imu.calcQuat(imu.qy), imu.calcQuat(imu.qz), aux_myRoll, myPitch, aux_myYaw);
  
//  SerialPort.println("Accel: " + String(accelX) + ", " +
//              String(accelY) + ", " + String(accelZ) + " g");
//  SerialPort.println();

  SerialPort.println(String(myRoll) + ", " +
              String(myPitch) + ", " + String(myYaw));
  SerialPort.println();

  // Send data over NRF
  NRFPacket.dAccelX = accelX; NRFPacket.dAccelY = accelY; NRFPacket.dAccelZ = accelZ;
  NRFPacket.dRoll = myRoll; NRFPacket.dPitch = myPitch; NRFPacket.dYaw = myYaw;
  radio.write(&NRFPacket, sizeof(NRFPacket));
}

